/* setting up the global program options */

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <gtk/gtk.h>
#include "main.h"

typedef struct {
	GtkWidget *okno ;
	GtkWidget *hist_button ;
	GtkWidget *file_button ;
	opt_s *o ;
} preferences_s ;


/* to reuse the code, default option vals + descripts are hold in entry_option_s */
entry_option_s options[] = {
	{ NULL, NULL, "command history items to store", "20", 1, NULL, 0.0, 20, 1, INT, NULL } ,
	{ NULL, NULL, "file history items to store", "5", 1, NULL, 0.0, 5, 1, INT, NULL } 
} ;


/* set up default option values on start */
int options_initialize(opt_s *opt) {

	opt->editable = FALSE ;
	opt->show_tooltips = TRUE ;
	opt->tips = gtk_tooltips_new() ;

	opt->font_name_default = my_strdup("-*-courier-medium-r-*-*-*-*-*-*-*-*-*-*") ;
	opt->font_name = opt->font_name_default ;
	opt->czcionka = gdk_font_load(opt->font_name_default) ;

	if(!opt->czcionka) 
		opt->czcionka = (gtk_widget_get_default_style())->font ;

	opt->input.on = TRUE ;
	opt->output.on = TRUE ;
	opt->error.on = TRUE ;

	opt->input.changed = FALSE ;
	opt->output.changed = FALSE ;
	opt->error.changed = FALSE ;

	opt->file_history = NULL ;

	opt->command_history = NULL ;
	opt->command_history = g_list_append(opt->command_history, "ls -l %input") ;
	opt->command_history = g_list_append(opt->command_history, "wc -c %input") ;

	opt->command_history_size = 20 ;
	opt->file_history_size = 5 ;

	opt->wdir = my_strdup(getenv("PWD")) ;
	opt->wdir = realloc(opt->wdir, strlen(opt->wdir) + 2000) ;
	strcat(opt->wdir, "/") ;

	if(opt->wdir != NULL) printf("Working directory %s\n", opt->wdir) ;


	/* we need temporary files for the gp programs to use */
	sprintf(opt->input.tmpfil,"/tmp/tescik-inputXXXXXX") ;
	sprintf(opt->output.tmpfil,"/tmp/tescik-outputXXXXXX") ;
	sprintf(opt->error.tmpfil,"/tmp/tescik-errorXXXXXX") ;

	if(mktemp(opt->input.tmpfil) == NULL ||
	  	mktemp(opt->output.tmpfil) == NULL ||
	  	mktemp(opt->error.tmpfil) == NULL) {
		komunikat("One or more of the temporary files could not be created.\n"
		 	"Since they are absolutely necessary, arka will now exit. Sorry.") ;
		exit(1) ;
	}

	/* each time a file is loaded, input gets erased; but error and output windows
	 * accumulate the produced data */
	opt->input.overwrite = TRUE ;
	opt->output.overwrite = FALSE ;
	opt->error.overwrite = FALSE ;

	opt->input.filenam = NULL ;
	opt->output.filenam = NULL ;
	opt->error.filenam = NULL ;

	opt->cursor_normal = gdk_cursor_new(GDK_ARROW) ;
	opt->cursor_wait = gdk_cursor_new(GDK_WATCH) ;

	return EXIT_SUCCESS ;
}


/* ---------------------------- preferences dialog ---------------------- */

/* closing the preferences dialog and storing the options */
void options_preferences_close(GtkWidget *kontr, gpointer dane) {
	preferences_s *pref ;
	opt_s * o ;

	pref = (preferences_s *) dane ;

	o = pref->o ;

	o->command_history_size =
		gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(options[0].widget)) ;
	o->file_history_size =
		gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(options[1].widget)) ;

	gtk_widget_destroy(pref->okno) ;
	
	gtk_main_quit() ;
	write_rcfile(o) ;
}


/* setting up the preferences dialog */
void options_preferences_set(gpointer dane, guint signal, GtkWidget *kontrolka) {
	GtkWidget *sep, *hpole, *ypole, *label, *notebook, *ok, *cancel ;
	preferences_s pref ;
	int nopts, i ;

	nopts = sizeof(options) / sizeof(entry_option_s) ;
  pref.o = (opt_s *) dane ;

	Warn("main preferences dialog") ;
	
	pref.okno = gtk_window_new(GTK_WINDOW_DIALOG) ;
	gtk_window_set_title(GTK_WINDOW(pref.okno), "Preferences") ;
	ypole = gtk_vbox_new(FALSE, 5) ;

	label = gtk_label_new("Main") ;

	notebook = gtk_notebook_new() ;
	gtk_notebook_append_page(GTK_NOTEBOOK(notebook), ypole, label) ;
	gtk_container_add(GTK_CONTAINER(pref.okno), notebook) ;

	/* updating options */
	options[0].intvalue = pref.o->command_history_size ;
	options[1].intvalue = pref.o->file_history_size ;

	/* same mechanism as in the programs menu -- to recycle the code */
	for(i = 0 ; i < nopts ; i++) {
		options[i].widget = set_widget_for_option(&options[i], ypole) ;
	}

	/* separator + OK + cancel buttons */
	sep = gtk_hseparator_new() ;
	gtk_box_pack_start(GTK_BOX(ypole), sep, FALSE, FALSE, 5) ;

	hpole = gtk_hbox_new(FALSE, 5) ;
	gtk_box_pack_start(GTK_BOX(ypole), hpole, FALSE, FALSE, 5) ;

	/* ok and cancel buttons */
	ok = gtk_button_new_with_label("Save") ;
	gtk_box_pack_start(GTK_BOX(hpole), ok, FALSE, FALSE, 10) ;
	gtk_signal_connect(GTK_OBJECT(ok), "clicked",
		GTK_SIGNAL_FUNC(options_preferences_close), &pref) ;

	cancel = gtk_button_new_with_label("Cancel") ;
	gtk_box_pack_start(GTK_BOX(hpole), cancel, FALSE, FALSE, 10) ;

	gtk_signal_connect(GTK_OBJECT(cancel), "clicked",
		GTK_SIGNAL_FUNC(zakoncz_dialog), pref.okno) ;

	GTK_WIDGET_SET_FLAGS(ok, GTK_CAN_DEFAULT) ;
	gtk_widget_grab_default(ok) ;
	gtk_widget_show_all(pref.okno) ;
	gtk_main() ;

}
